@@ -55,6 +55,101 @@ class Insight_Definition:
             self.app_url = "https://app.askdata.com/"
 
 
+    def add_widget(self, widget_type: str, query_ids:list=[], settings:dict={}, texts:dict={}, charts:list=[]):
+        position = (len(self.components))
+        widget_id = self.__create_widget(position, widget_type)
+        if(query_ids!=[]):
+            self.edit_widget(widget_id, widget_type,query_ids, settings, texts, charts)
+        return widget_id
+
+    def edit_widget(self, widget_id, widget_type: str, query_ids:list, field_codes:dict, texts:dict={}, charts:list=[]):
+
+        url = self.smart_insight_url+"/definitions/"+self.definition_id+"/widgets/"+widget_id
+
+        columns = self.__get_columns(query_ids)
+        settings = {}
+        for field in field_codes.keys():
+            for column in columns:
+                if(column["code"] in field_codes[field]):
+                    for query in query_ids:
+                        if(str(field_codes[field]).startswith(query)):
+                           if(str(column["id"]).startswith(query)):
+                               settings[field + "_id"] = column["id"]
+                               break
+                           else:
+                               continue
+                        else:
+                            settings[field + "_id"] = column["id"]
+                            break
+
+        logging.info("SETTINGS {}".format(settings))
+        body = {
+            "id" : widget_id,
+            "type": "widget",
+            "name": "Widget",
+            "customName": False,
+            "dependsOn": [],
+            "widgetType": widget_type,
+            "queryIds": query_ids,
+            "settings": settings,
+            "texts": texts,
+            "charts": charts,
+            "valid": True,
+            "queryComponent": False
+        }
+
+        s = requests.Session()
+        s.keep_alive = False
+        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
+        s.mount('https://', HTTPAdapter(max_retries=retries))
+
+        headers = {
+            "Content-Type": "application/json",
+            "Authorization": "Bearer" + " " + self._token
+        }
+        r = s.put(url=url, json=body, headers=headers)
+        r.raise_for_status()
+        self.components = r.json()["components"]
+
+    def __get_columns(self, query_ids) -> list:
+
+        url = self.smart_insight_url+"/definitions/"+self.definition_id+"/queries/columns"
+        body = {"queryIds": query_ids}
+        s = requests.Session()
+        s.keep_alive = False
+        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
+        s.mount('https://', HTTPAdapter(max_retries=retries))
+
+        headers = {
+            "Content-Type": "application/json",
+            "Authorization": "Bearer" + " " + self._token
+        }
+        logging.info("URL {}".format(url))
+        r = s.post(url=url, json=body, headers=headers)
+        r.raise_for_status()
+        return r.json()
+
+    def __create_widget(self, position, widget_type):
+
+        body = {"type": "widget", "position": position, "widgetType": widget_type}
+
+        s = requests.Session()
+        s.keep_alive = False
+        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
+        s.mount('https://', HTTPAdapter(max_retries=retries))
+
+        headers = {
+            "Content-Type": "application/json",
+            "Authorization": "Bearer" + " " + self._token
+        }
+        query_url = self.smart_insight_url + '/definitions/' + self.definition_id + '/components/'
+        logging.info("URL {}".format(query_url))
+        r = s.post(url=query_url, json=body, headers=headers)
+        r.raise_for_status()
+        print(r.json())
+        self.components = r.json()["components"]
+        return self.components[position]["id"]
+
     def add_table(self, query="", columns=[]):
 
         position = (len(self.components))
@@ -1073,3 +1168,27 @@ class Insight_Definition:
 
         r = s.post(url=url, json=body, headers=headers)
         r.raise_for_status()
+
+    def clone(self, channel_id, card_name=""):
+
+        if card_name == "":
+            card_name = "Copy of " + self.name
+        url = "https://api-dev.askdata.com/smartinsight/definitions/1dae1fcd-d96f-4c9f-ba7e-cdab6b34ff1d/clone"
+        body = {
+            "agentId": self.agent_id,
+            "name": card_name,
+            "channelId": channel_id
+            }
+
+        s = requests.Session()
+        s.keep_alive = False
+        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
+        s.mount('https://', HTTPAdapter(max_retries=retries))
+
+        headers = {
+            "Content-Type": "application/json",
+            "Authorization": "Bearer" + " " + self._token
+        }
+
+        r = s.post(url=url, json=body, headers=headers)
+        r.raise_for_status()
\ No newline at end of file
